<?php

namespace Drupal\xnttbrapi\Plugin\ExternalEntities\StorageClient;

use Drupal\Core\Form\FormStateInterface;
use Drupal\external_entities\Plugin\ExternalEntities\StorageClient\Rest;
use Drupal\xnttbrapi\Exception\XnttBrapiException;
use GuzzleHttp\Exception\ClientException;
use GuzzleHttp\Exception\ConnectException;
use GuzzleHttp\Exception\RequestException;
use GuzzleHttp\Exception\ServerException;

/**
 * External entities storage client based on a REST API.
 *
 * @ExternalEntityStorageClient(
 *   id = "brapi",
 *   label = @Translation("BrAPI"),
 *   description = @Translation("Retrieves external entities from a Breeding API enpoint.")
 * )
 */
class Brapi extends Rest {

  /**
   * {@inheritdoc}
   */
  public function defaultConfiguration() {
    $config = parent::defaultConfiguration();
    $config['endpoint_search'] = NULL;
    return $config;
  }

  /**
   * {@inheritdoc}
   */
  public function buildConfigurationForm(
    array $form,
    FormStateInterface $form_state
  ) {
    $form = parent::buildConfigurationForm($form, $form_state);

    // Add help.
    $form['endpoint']['#title'] = $this->t('BrAPI Call URL');
    $form['endpoint']['#description'] = $this->t(
      'Enter the BrAPI call URL including curly brackets with object identifier. Ex.:<ul><li>"https://www.crop-diversity.org/mgis/brapi/v1/germplasm/{germplasmDbId}"</li><li>"https://www.crop-diversity.org/mgis5/brapi/v2/germplasm?accessionNumber={accessionNumber}".</li></ul>Don\'t forget to map the external entity identifier to the field named inside the curly brackets.'
    );
    $form['api_key']['key']['#description'] = $this->t(
      'Enter here the BrAPI token to use if the endpoint requires one.'
    );
    $form['api_key']['key']['#title'] = $this->t('BrAPI Token');
    // Remove 'Bearer ' part automatically added by form validation.
    $form['api_key']['key']['#default_value'] = str_replace(
      'Bearer ',
      '',
      $this->configuration['api_key']['key'] ?? ''
    );

    // Auto-set some REST parameters.
    $form['response_format'] = [
      '#type' => 'hidden',
      '#value' => 'json',
    ];

    $form['pager']['page_parameter'] = [
      '#type' => 'hidden',
      '#value' => 'page',
    ];

    $form['pager']['page_parameter_type'] = [
      '#type' => 'hidden',
      '#value' => 'pagenum',
    ];

    $form['pager']['page_size_parameter'] = [
      '#type' => 'hidden',
      '#value' => 'pageSize',
    ];

    $form['pager']['page_size_parameter_type'] = [
      '#type' => 'hidden',
      '#value' => 'pagesize',
    ];

    $form['api_key']['header_name'] = [
      '#type' => 'hidden',
      '#value' => 'Authorization',
    ];

    $form['parameters']['#type'] = 'details';
    $form['parameters']['#title'] = $this->t('Optional parameters');
    $form['parameters']['#open'] = FALSE;
    $form['parameters']['#description'] = $this->t(
      'Optional query parameters can be specified to pre-filter results. For instance, use "instituteCode|TheInstituteName" to restrict germplasm to the institute "TheInstituteName" or "species|SomeSpeciesName" to restrict to the species "SpeciesName".'
    );

    $form['parameters']['single'] = [
      '#type' => 'hidden',
      '#value' => '',
    ];

    $form['parameters']['list']['#title'] = $this->t('URL query parameters');
    $form['parameters']['list']['#description'] = '';

    // Add search call.
    $form['endpoint_search'] = [
      '#type' => 'textfield',
      '#title' => $this->t('Search call'),
      '#required' => FALSE,
      '#description' => $this->t('Optional but recommended. Search call corresponding to the given BrAPI call to faster retrieve multiple objects in one (or few) POST call(s). Both the full URL or just the call name are supported (ie.: "search/germplasm"). Leave empty if not available.'),
      '#default_value' => $this->configuration['endpoint_search'],
    ];

    return $form;
  }

  /**
   * {@inheritdoc}
   */
  public function validateConfigurationForm(
    array &$form,
    FormStateInterface $form_state
  ) {
    // Prefix key if needed.
    $api_key = $form_state->getValue('api_key');
    if (!empty($api_key['key'])
        && (@substr_compare($api_key['key'], 'Bearer', 0, 6) !== 0)
    ) {
      $api_key['key'] = 'Bearer ' . $api_key['key'];
      $form_state->setValue('api_key', $api_key);
    }

    parent::validateConfigurationForm($form, $form_state);
  }

  /**
   * Returns the BrAPI call URL.
   *
   * @param string $id
   *   A BrAPI entity identifier or an empty string if none for listing.
   *
   * @return string
   *   The BrAPI endpoint URL qualified with the given identifier if one.
   */
  public function getEndpoint(string $id = '') :string {
    $brapi_url = $this->configuration['endpoint'];
    // Add the entity identifier to the URL.
    if ($placholder_count = preg_match_all('#(\{\w+\})#', $brapi_url, $matches)) {
      // Use specified placeholder if one.
      if (empty($id)) {
        $brapi_url = preg_replace(
          '#(?:[^/?&=;]+=|/)?\{\w+\}.*#',
          '',
          $brapi_url
        );
      }
      elseif (1 < $placholder_count) {
        // Manage multiple placeholders and identifiers, like
        // "maps/{mapDbId}/positions/{linkageGroupName}" using '/' as a
        // separator in the given identifier to split it into multiple
        // identifiers.
        // First, check if the given id matches the format.
        $ids = array_filter(explode('/', $id), 'strlen');
        if (count($ids) == $placholder_count) {
          // Got the correct number of identifiers, replace each.
          foreach ($ids as $index => $sub_id) {
            $placeholder = $matches[1][$index];
            $pos = strpos($brapi_url, $placeholder);
            if ($pos !== false) {
              $brapi_url = substr_replace(
                $brapi_url,
                $sub_id,
                $pos,
                strlen($placeholder)
              );
            }
          }
        }
        else {
          // Invalid count, replace all with an invalid id.
          $brapi_url = preg_replace('#\{\w+\}#', '-', $brapi_url);
        }
      }
      else {
        $brapi_url = preg_replace('#\{\w+\}#', $id, $brapi_url);
      }
    }
    else {
      // Append otherwise.
      $brapi_url = rtrim($brapi_url, '/') . '/' . $id;
    }

    return $brapi_url;
  }

  /**
   * Returns the provided BrAPI search call URL if one.
   *
   * @return string
   *   The BrAPI endpoint search URL if one or an empty string.
   */
  public function getSearchCall() :string {
    $search_url = trim(
      $this->configuration['endpoint_search'] ?? '',
      " \n\r\t\v/"
    );
    // Check URL format.
    if (!empty($search_url)) {
      if (FALSE === strripos($search_url, 'search')) {
        \Drupal::logger('xnttbrapi')->warning(
          'The call provided for search call ({call}) does not appear to be a BrAPI search call. Check the settings.',
          ['call' => $search_url]
        );
      }
      if ((0 !== strncmp('http', $search_url, 4))
          && (':' != $search_url[0])
      ) {
        // Not a full URL.
        if (preg_match('#^(.*/brapi/v\d/)#', $this->getEndpoint(), $matches)) {
          $search_url = $matches[1] . $search_url;
        }
        else {
          // Failed to parse URL.
          \Drupal::logger('xnttbrapi')->warning(
            'Failed to parse BrAPI base URL for search call. Try to provide a full BrAPI search call URL in settings.'
          );
          $search_url = '';
        }
      }
    }
    return $search_url;
  }

  /**
   * Returns the BrAPI object identifier used.
   *
   * @return string
   *   The BrAPI identifier field name or throws an excetpion.
   *
   * @throw XnttBrapiException if unable to guess object identifier.
   */
  public function getIdentifierFieldName() :string {
    $call_url = $this->configuration['endpoint'];
    if (!preg_match('#\{(\w+)\}#', $call_url, $matches) || empty($matches[1])) {
      throw new XnttBrapiException(
        "Error: unable to guess identifier field from call ($call_url). No field specified in curly braces."
      );
    }
    return $matches[1];
  }

  /**
   * {@inheritdoc}
   */
  public function loadMultiple(array $ids = NULL) {
    $data = [];

    if (!empty($ids) && is_array($ids)) {
      // Check if a search call is available.
      $search_url = $this->getSearchCall();
      if (!empty($search_url)) {
        // Load multiple entities at once using search call.
        // Support query string in provided BrAPI endpoint.
        list($search_url, $brapi_query) = explode('?', $search_url) + [1 => ''];
        // Note: $search_url returned by getSearchCall() cannot start by '?'.
        if ('/' != $search_url[-1]) {
          $search_url .= '/';
        }
        if (!empty($brapi_query)) {
          // Note: we don't consider ';' as a query parameter separator as it is not
          // used by BrAPI specifications.
          preg_match_all('#([^&=]+)(?:=([^&=]*))?#', $brapi_query, $matches);
          $brapi_query = array_combine($matches[1], $matches[2]);
        }
        else {
          $brapi_query = [];
        }
        // Endpoint parameters come first.
        $query = $brapi_query + $this->getPagingQueryParameters();
        $method = '';
        $format = $this->configuration['response_format'];
        $decoder = $this->getResponseDecoderFactory()->getDecoder($format);
        try {
          // Get identifier field name.
          $id_field = $this->getIdentifierFieldName();
          $method = 'POST';
          $request = [
            'headers' => $this->getHttpHeaders(),
            'query' => $query,
            'json' => [$id_field . 's' => $ids, $id_field => $ids, ],
          ];
          $response = $this->httpClient->request(
            $method,
            $search_url,
            $request
          );

          // Manage asynchronous searches with "searchResultsDbId".
          if (202 == $response->getStatusCode()) {
            // Asynchronous search, get search token.
            $body = ltrim($response->getBody(), "\xEF\xBB\xBF");
            $result = $decoder->decode($body);
            $search_results_db_id = $result['result']['searchResultsDbId'] ?? '';
            if (empty($search_results_db_id)) {
              throw new XnttBrapiException(
                "Failed to use BrAPI search call:\nURL: $search_url\nMethod: $method\nQuery: "
                . print_r($query, TRUE)
                . 'Error: no searchResultsDbId returned while 202 HTTP response code.'
              );
            }
            $method = 'GET';
            $search_url = $search_url . '/' . $search_results_db_id;
            $request = ['headers' => $request['headers'], ];
            // Try for 5 min max.
            $timeout = time() + 300;
            $trials = 0;
            // Start to query every 1 sec.
            $sleep_time = 1;
            while ((time() < $timeout)
              && (202 == $response->getStatusCode())
            ) {
              // Wait before re-querying.
              sleep($sleep_time);
              switch (++$trials) {
                case 10:
                  // Switch to 2sec delay after 10sec.
                  $sleep_time = 2;
                  break;
                case 15:
                  // Switch to 5sec delay after 20sec.
                  $sleep_time = 5;
                  break;
                case 29:
                  // Switch to 10sec delay after 1min30.
                  $sleep_time = 10;
                  break;
              }
              $response = $this->httpClient->request(
                $method,
                $search_url,
                $request
              );
            }
          }
          if (200 != $response->getStatusCode()) {
            // Queery failed.
            throw new XnttBrapiException(
              "Failed to use BrAPI search call:\nURL: $search_url\nMethod: $method\nQuery: "
              . print_r($query, TRUE)
              . 'Trials: '
              . ($trials+1)
              . "\nError: Got HTTP response code "
              . $response->getStatusCode()
            );
          }
          // Manage results and pagination.
          do {
            // Remove BOM.
            $body = ltrim($response->getBody(), "\xEF\xBB\xBF");
            $result = $decoder->decode($body);
            $total_pages =
              $result['metadata']['pagination']['totalPages']
              ?? NULL
            ;
            $current_page =
              $result['metadata']['pagination']['currentPage']
              ?? NULL
            ;
            if (!isset($total_pages) || !isset($current_page)) {
              // Pagination failed.
              throw new XnttBrapiException(
                "Failed to use BrAPI search call:\nURL: $search_url\nMethod: $method\nQuery: "
                . print_r($query, TRUE)
                . "\nError: Missing pagination data in response:\n"
                . print_r($result, TRUE)
              );
            }
            // To work with search calls or GET filters, we need to support one
            // object returned into the 'data' key.
            if (isset($result['result']['data'])) {
              // Result data array is set, does it contain something?
              if (!empty($result['result']['data'])) {
                $missing_ids = 0;
                foreach ($result['result']['data'] as $brapi_object) {
                  if (isset($brapi_object[$id_field])) {
                    $data[$brapi_object[$id_field]] = $brapi_object;
                  }
                  else {
                    // Warn about data issue.
                    ++$missing_ids;
                  }
                }
                if ($missing_ids) {
                  \Drupal::logger('xnttbrapi')->warning(
                    "Warning when querying BrAPI endpoint:\nURL: $search_url\nMethod: $method\nQuery: "
                    . print_r($query, TRUE)
                    . 'Warning: Failed to get '
                    . $missing_ids
                    . ' object(s) because of missing identifiers ("'
                    . $id_field
                    .'" field)'
                    . $response_body
                  );
                }
                // Check for invalid BrAPI implementation that does not filter.
                if (count($data) > count($ids)) {
                  throw new XnttBrapiException(
                    "Failed to use BrAPI search call:\nURL: $search_url\nMethod: $method\nQuery: "
                    . print_r($query, TRUE)
                    . "\nJSON input: "
                    . print_r([$id_field => $ids, ], TRUE)
                    . "\nError: The search call returns more entries than requested. It seems the filtering is not applied properly. Check the BrAPI call behavior and supported features."
                  );
                }
              }
            }
            elseif (!array_key_exists('result', $result)
              || (!empty($result['result']))
            ) {
              // We don't go here if 'result' is just empty, which is ok.
              if (!empty($result['result']['searchResultsDbId'])) {
                // BrAPI specifications not followed for asynchronous search.
                throw new XnttBrapiException(
                  "Failed to use BrAPI search call:\nURL: $search_url\nMethod: $method\nQuery: "
                  . print_r($query, TRUE)
                  . "\nError: The search call uses asyncronous search but does not return the 202 HTTP status code. Not BrAPI-compliant."
                );
              }
              // Not the expected result format for a search call.
              throw new XnttBrapiException(
                "Failed to use BrAPI search call:\nURL: $search_url\nMethod: $method\nQuery: "
                . print_r($query, TRUE)
                . "\nError: Unexpected data structure in response:\n"
                . print_r($result, TRUE)
              );
            }

            // Proceed next page.
            ++$current_page;
            if ($current_page < $total_pages) {
              $query['page'] = $current_page;
              $request['query'] = $query;
              $response = $this->httpClient->request(
                $method,
                $search_url,
                $request
              );
            }
          } while ($current_page < $total_pages);
          // Done.
        }
        catch (ServerException $e) {
          \Drupal::messenger()->addWarning(
            $this->t('Failed to query BrAPI endpoint. See logs for details.')
          );
          $response = $e->getResponse();
          $response_body = $response->getBody()->getContents();
          \Drupal::logger('xnttbrapi')->warning(
            "Failed to query BrAPI endpoint:\nURL: $search_url\nMethod: $method\nQuery: "
            . print_r($query, TRUE)
            . 'Error: '
            . $response_body
          );
        }
        catch (ClientException $e) {
          // Not found or some other 4xx error.
        }
        catch (ConnectException | RequestException $e) {
          \Drupal::messenger()->addWarning(
            $this->t('Failed to connect to BrAPI endpoint. See logs for details.')
          );
          \Drupal::logger('xnttbrapi')->warning(
            "Failed to connect to BrAPI endpoint:\nURL: $search_url\nMethod: $method\nQuery: "
            . print_r($query, TRUE)
            . 'Error: '
            . $e
          );
        }
        catch (XnttBrapiException $e) {
          \Drupal::messenger()->addWarning(
            $this->t('Failed to use BrAPI search call "' . $search_url . '". See logs for details.')
          );
          \Drupal::logger('xnttbrapi')->warning(
            '{message}',
            ['message' => $e->getMessage(), ]
          );
        }
      }
      else {
        // Slow mode; load one by one...
        foreach ($ids as $id) {
          $data[$id] = $this->load($id);
        }
      }
    }

    return $data;
  }

  /**
   * {@inheritdoc}
   */
  public function load($id) {

    $brapi_url = $this->getEndpoint($id);
    // Support query string in provided BrAPI endpoint.
    list($brapi_url, $brapi_query) = explode('?', $brapi_url) + [1 => ''];
    if (!empty($brapi_query)) {
      // Note: we don't consider ';' as a query parameter separator as it is not
      // used by BrAPI specifications.
      preg_match_all('#([^&=]+)(?:=([^&=]*))?#', $brapi_query, $matches);
      $brapi_query = array_combine($matches[1], $matches[2]);
    }
    else {
      $brapi_query = [];
    }
    // Endpoint parameters come first (object ID won't be overriden).
    $query = $brapi_query + $this->getPagingQueryParameters();
    try {
      $response = $this->httpClient->request(
        'GET',
        $brapi_url,
        [
          'headers' => $this->getHttpHeaders(),
          'query' => $query,
        ]
      );

      $format = $this->configuration['response_format'];
      // Remove BOM.
      $body = ltrim($response->getBody(), "\xEF\xBB\xBF");

      $result = $this
        ->getResponseDecoderFactory()
        ->getDecoder($format)
        ->decode($body)
      ;
    }
    catch (ServerException $e) {
      \Drupal::messenger()->addWarning(
        $this->t('Failed to query BrAPI endpoint. See logs for details.')
      );
      $response = $e->getResponse();
      $response_body = $response->getBody()->getContents();
      \Drupal::logger('xnttbrapi')->warning(
        "Failed to query BrAPI endpoint:\nURL: $brapi_url\nMethod: GET\nQuery: "
        . print_r($query, TRUE)
        . 'Error: '
        . $response_body
      );
    }
    catch (ClientException $e) {
      // Not found or some other 4xx error.
    }
    catch (ConnectException | RequestException $e) {
      \Drupal::messenger()->addWarning(
        $this->t('Failed to connect to BrAPI endpoint. See logs for details.')
      );
      \Drupal::logger('xnttbrapi')->warning(
        "Failed to connect to BrAPI endpoint:\nURL: $brapi_url\nMethod: GET\nQuery: "
        . print_r($query, TRUE)
        . 'Error: '
        . $e
      );
    }
    // To work with search calls or GET filters, we need to support one object
    // returned into the 'data' key.
    if (isset($result['result']['data'])) {
      if (!empty($result['result']['data'])) {
        $data = $result['result']['data'][0];
        // More than one object means the filter did not work to return a single
        // object as expected.
        if (1 < count($result['result']['data'])) {
          \Drupal::logger('xnttbrapi')->warning(
            'The given endpoint returned more than one object for ID "'
            . $id
            . '" from: '
            . $this->getEndpoint($id)
          );
        }
      }
      else {
        $data = NULL;
      }
    }
    else {
      // Otherwise, regular call result: the object is into the "result" key.
      $data = $result['result'] ?? NULL;
    }
    return $data;
  }

  /**
   * {@inheritdoc}
   */
  public function query(
    array $parameters = [],
    array $sorts = [],
    $start = NULL,
    $length = NULL
  ) {
    // Note: the default $length to 50 is set by ExternalEntityListBuilder (hard
    // coded) so the default number of item per page is not taken into account
    // on regular external entity listing pages.
    // See https://www.drupal.org/project/external_entities/issues/3136353.

    $brapi_url = $this->getEndpoint();
    // Support query string in provided BrAPI endpoint.
    list($brapi_url, $brapi_query) = explode('?', $brapi_url) + [1 => ''];
    if (!empty($brapi_query)) {
      // Note: we don't consider ';' as a query parameter separator as it is not
      // used by BrAPI specifications.
      preg_match_all('#([^&=]+)(?:=([^&=]*))?#', $brapi_query, $matches);
      $brapi_query = array_combine($matches[1], $matches[2]);
    }
    else {
      $brapi_query = [];
    }
    // Pager and list parameters come first (discard endpoint parameter
    // overrides for listing).
    $query =
      $this->getListQueryParameters($parameters, $start, $length)
      + $brapi_query
    ;
    try {
      $response = $this->httpClient->request(
        'GET',
        $this->getEndpoint(),
        [
          'headers' => $this->getHttpHeaders(),
          'query' => $query,
        ]
      );

      $format = $this->configuration['response_format'];
      $body = ltrim($response->getBody(), "\xEF\xBB\xBF");

      $results = $this
        ->getResponseDecoderFactory()
        ->getDecoder($format)
        ->decode($body);
    }
    catch (ServerException | ClientException $e) {
      \Drupal::messenger()->addWarning(
        $this->t('Failed to query BrAPI endpoint. See logs for details.')
      );
      $response = $e->getResponse();
      $response_body = $response->getBody()->getContents();
      \Drupal::logger('xnttbrapi')->warning(
        "Failed to query BrAPI endpoint:\nURL: $brapi_url\nMethod: GET\nQuery: "
        . print_r($brapi_query, TRUE)
        . 'Error: '
        . $response_body
      );
    }
    catch (ConnectException | RequestException $e) {
      \Drupal::messenger()->addWarning(
        $this->t('Failed to connect to BrAPI endpoint. See logs for details.')
      );
      \Drupal::logger('xnttbrapi')->warning(
        'Failed to connect to BrAPI endpoint: '
        . $e
      );
    }

    $items = [];
    if (!empty($results['result']['data'])) {
      $items = array_values($results['result']['data']);
    }

    return $items;
  }

  /**
   * {@inheritdoc}
   */
  public function countQuery(array $parameters = []) {
    $brapi_url = $this->getEndpoint();
    // Support query string in provided BrAPI endpoint.
    list($brapi_url, $brapi_query) = explode('?', $brapi_url) + [1 => ''];
    if (!empty($brapi_query)) {
      // Note: we don't consider ';' as a query parameter separator as it is not
      // used by BrAPI specifications.
      preg_match_all('#([^&=]+)(?:=([^&=]*))?#', $brapi_query, $matches);
      $brapi_query = array_combine($matches[1], $matches[2]);
    }
    else {
      $brapi_query = [];
    }
    // Pager and list parameters come first (discard endpoint parameter
    // overrides for listing).
    $query = $this->getListQueryParameters($parameters, 0, 1) + $brapi_query;
    try {
      $response = $this->httpClient->request(
        'GET',
        $brapi_url,
        [
          'headers' => $this->getHttpHeaders(),
          'query' => $query,
        ]
      );

      $format = $this->configuration['response_format'];
      $body = ltrim($response->getBody(), "\xEF\xBB\xBF");

      $results = $this
        ->getResponseDecoderFactory()
        ->getDecoder($format)
        ->decode($body);
    }
    catch (ServerException | ClientException $e) {
      \Drupal::messenger()->addWarning(
        $this->t('Failed to count BrAPI endpoint. See logs for details.')
      );
      $response = $e->getResponse();
      $response_body = $response->getBody()->getContents();
      \Drupal::logger('xnttbrapi')->warning(
        'Failed to count BrAPI endpoint: '
        . $response_body
      );
    }
    catch (ConnectException | RequestException $e) {
      \Drupal::messenger()->addWarning(
        $this->t('Failed to connect to BrAPI endpoint. See logs for details.')
      );
      \Drupal::logger('xnttbrapi')->warning(
        'Failed to connect to BrAPI endpoint: '
        . $e
      );
    }

    $count = 0;
    // Tries to get data count from the API.
    if (!empty($results['metadata']['pagination']['totalCount'])) {
      $count = $results['metadata']['pagination']['totalCount'];
    }
    else {
      // Fallback to default if it failed.
      $count = parent::countQuery($parameters);
    }

    return $count;
  }
}
